#lang plai

(define-type FAE
  [num (n number?)]
  [add (lhs FAE?)
       (rhs FAE?)]
  [sub (lhs FAE?)
       (rhs FAE?)]
  [id (name symbol?)]
  [fun (param symbol?)
       (body FAE?)]
  [app (fun-expr FAE?)
       (arg-expr FAE?)]
  [if0 (test FAE?)
       (then FAE?)
       (else FAE?)])

(define-type CFAE
  [cnum (n number?)]
  [cadd (lhs CFAE?)
        (rhs CFAE?)]
  [csub (lhs CFAE?)
        (rhs CFAE?)]
  [cid (pos number?)]
  [cfun (body CFAE?)]
  [capp (fun-expr CFAE?)
       (arg-expr CFAE?)]
  [cif0 (test CFAE?)
        (then CFAE?)
        (else CFAE?)])

(define-type CFAE-Value
  [numV (n number?)]
  [closureV (body CFAE?)
            (ds DefrdSub?)])

(define-type CDefrdSub
  [mtCSub]
  [aCSub (name symbol?)
         (rest CDefrdSub?)])

(define-type CFAE-Cont
  [mtK]
  [addSecondK (r CFAE?)
              (ds DefrdSub?)
              (k CFAE-Cont?)]
  [doAddK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
  [subSecondK (r CFAE?)
              (ds DefrdSub?)
              (k CFAE-Cont?)]
  [doSubK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
  [appArgK (arg-expr CFAE?)
           (ds DefrdSub?)
           (k CFAE-Cont?)]
  [doAppK (fun-val CFAE-Value?)
          (k CFAE-Cont?)]
  [doIfK (then-expr CFAE?)
         (else-expr CFAE?)
         (ds DefrdSub?)
         (k CFAE-Cont?)])

;; A DefrdSub is a list of CFAE-Value
(define (DefrdSub? x) true)

;; ----------------------------------------

;; parse : S-expr -> FAE
(define (parse sexp)
  (cond
    [(number? sexp) (num sexp)]
    [(symbol? sexp) (id sexp)]
    [(pair? sexp)
     (case (first sexp)
       [(+) (add (parse (second sexp)) (parse (third sexp)))]
       [(-) (sub (parse (second sexp)) (parse (third sexp)))]
       [(fun) (fun (first (second sexp)) (parse (third sexp)))]
       [(if0) (if0 (parse (second sexp))
                   (parse (third sexp))
                   (parse (fourth sexp)))]
       [else (app (parse (first sexp)) (parse (second sexp)))])]))

(test (parse 3) (num 3))
(test (parse 'x) (id 'x))
(test (parse '{+ 1 2}) (add (num 1) (num 2)))
(test (parse '{- 1 2}) (sub (num 1) (num 2)))
(test (parse '{fun {x} x}) (fun 'x (id 'x)))
(test (parse '{1 2}) (app (num 1) (num 2)))
(test (parse '{if0 0 1 2}) (if0 (num 0) (num 1) (num 2)))

;; ----------------------------------------

;; compile : FAE CDefrdSub -> CFae
(define (compile a-fae ds)
  (type-case FAE a-fae
    [num (n) (cnum n)]
    [add (l r) (cadd (compile l ds) (compile r ds))]
    [sub (l r) (csub (compile l ds) (compile r ds))]
    [id (name) (cid (locate name ds))]
    [fun (param body-expr) (cfun 
                            (compile body-expr (aCSub param ds)))]
    [app (fun-expr arg-expr)
         (capp (compile fun-expr ds)
               (compile arg-expr ds))]
    [if0 (test-expr then-expr else-expr)
         (cif0 (compile test-expr ds)
               (compile then-expr ds)
               (compile else-expr ds))]))

(define (locate name ds)
  (type-case CDefrdSub ds
    [mtCSub () (error 'locate "free variable")]
    [aCSub (sub-name rest-ds)
           (if (symbol=? sub-name name)
               0
               (+ 1 (locate name rest-ds)))]))

;; ----------------------------------------

;; interp : FAE DefrdSub FAE-Cont -> FAE-Value
(define (interp a-fae ds k)
  (type-case CFAE a-fae
    [cnum (n) (continue k (numV n))]
    [cadd (l r) (interp l ds (addSecondK r ds k))]
    [csub (l r) (interp l ds (subSecondK r ds k))]
    [cid (pos) (continue k (list-ref ds pos))]
    [cfun (body-expr)
          (continue k (closureV body-expr ds))]
    [capp (fun-expr arg-expr)
          (interp fun-expr ds (appArgK arg-expr ds k))]
    [cif0 (test-expr then-expr else-expr)
          (interp test-expr ds (doIfK then-expr else-expr ds k))]))

;; continue : FAE-Cont FAE-Value -> FAE-Value
(define (continue k v)
  (type-case CFAE-Cont k
    [mtK () v]
    [addSecondK (r ds k)
                (interp r ds (doAddK v k))]
    [doAddK (v1 k)
            (continue k (num+ v1 v))]
    [subSecondK (r ds k)
                (interp r ds (doSubK v k))]
    [doSubK (v1 k)
            (continue k (num- v1 v))]
    [appArgK (arg-expr ds k)
             (interp arg-expr ds (doAppK v k))]
    [doAppK (fun-val k)
            (interp (closureV-body fun-val)
                    (cons v
                          (closureV-ds fun-val))
                    k)]
    [doIfK (then-expr else-expr ds k)
           (if (numzero? v)
               (interp then-expr ds k)
               (interp else-expr ds k))]))

;; num-op : (number number -> number) -> (FAE-Value FAE-Value -> FAE-Value)
(define (num-op op op-name)
  (lambda (x y)
    (numV (op (numV-n x) (numV-n y)))))

(define num+ (num-op + '+))
(define num- (num-op - '-))

;; numzero? : FAE-Value -> boolean
(define (numzero? x)
  (zero? (numV-n x)))

;; ----------------------------------------

(define init-k (mtK))

(test (interp (compile (num 10) (mtCSub))
              empty
              init-k)
      (numV 10))
(test (interp (compile (add (num 10) (num 7)) (mtCSub))
              empty
              init-k)
      (numV 17))
(test (interp (compile (sub (num 10) (num 7)) (mtCSub))
              empty
              init-k)
      (numV 3))
(test (interp (compile
               (app (fun 'x (add (id 'x) (num 12)))
                    (add (num 1) (num 17)))
               (mtCSub))
              empty
              init-k)
      (numV 30))
(test (interp (compile (id 'x)
                       (aCSub 'x (mtCSub)))
              (list (numV 10))
              init-k)
      (numV 10))

(test (interp (compile (app (fun 'x (add (id 'x) (num 12)))
                            (add (num 1) (num 17)))
                       (mtCSub))
              empty
              init-k)
      (numV 30))

(test (interp (compile (app (fun 'x
                                 (app (fun 'f
                                           (add (app (id 'f) (num 1))
                                                (app (fun 'x
                                                          (app (id 'f) (num 2)))
                                                     (num 3))))
                                      (fun 'y (add (id 'x) (id 'y)))))
                            (num 0))
                       (mtCSub))
              empty
               init-k)
      (numV 3))

(test (interp (compile (if0 (num 0)
                            (num 1)
                            (num 2))
                       (mtCSub))
              empty
               init-k)
      (numV 1))
(test (interp (compile (if0 (num 1)
                            (num 0)
                            (num 2))
                       (mtCSub))
              empty
              init-k)
      (numV 2))


(test (interp (compile
               (parse 
                '{{fun {mkrec}
                       {{fun {fib}
                             ;; Call fib on 4:
                             {fib 4}}
                        ;; Create recursive fib:
                        {mkrec
                         {fun {fib}
                              ;; Fib:
                              {fun {n}
                                   {if0 n
                                        1
                                        {if0 {- n 1}
                                             1
                                             {+ {fib {- n 1}}
                                                {fib {- n 2}}}}}}}}}}
                  ;; mkrec:
                  {fun {body-proc}
                       {{fun {fX}
                             {fX fX}}
                        {fun {fX}
                             {body-proc {fun {x} {{fX fX} x}}}}}}})
               (mtCSub))
              empty
              init-k)
      (numV 5))

(test/exn (compile (id 'x) (mtCSub))
          "free variable")
